
// sn76489_clock : 3579545
// dcsg f-number : clock / 32 / freq

#include "dcsgdrv.hpp"

static const unsigned short DcsgKeyTable[] = {
	0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,	// o0
	0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,	// o1
	0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03ff,0x03f8,0x03bf,0x0389,	// o2
	0x0357,0x0327,0x02f9,0x02cf,0x02a6,0x0280,0x025c,0x023a,0x021a,0x01fc,0x01df,0x01c4,	// o3
	0x01ab,0x0193,0x017c,0x0167,0x0153,0x0140,0x012e,0x011d,0x010d,0x00fe,0x00ef,0x00e2,	// o4
	0x00d5,0x00c9,0x00be,0x00b3,0x00a9,0x00a0,0x0097,0x008e,0x0086,0x007f,0x0077,0x0071,	// o5
	0x006a,0x0064,0x005f,0x0059,0x0054,0x0050,0x004b,0x0047,0x0043,0x003f,0x003b,0x0038,	// o6
	0x0035,0x0032,0x002f,0x002c,0x002a,0x0028,0x0025,0x0023,0x0021,0x001f,0x001d,0x001c,	// o7
	0x001a,0x0019,0x0017,0x0016,0x0015,0x0014,0x0012,0x0011,0x0010,0x000f,0x000e,0x000e,	// o8
};

static const unsigned int freq_slot[] = { 0,2,4,6 };
static const unsigned int volume_slot[] = { 1,3,5,7 };

//---------------------------------------------------------------------

void
DCSGDRV::SetReg(unsigned int value)
{
	if (pVgmOut) {
		pVgmOut->WriteRegDcsg(value);
	}
}

//---------------------------------------------------------------------

DCSGDRV::DCSGDRV()
{
	pVgmOut = NULL;
	for (unsigned int i = 0; i < DCSG_CHANNEL_MAX; i++) {
		Channel[i].key_on = false;
		Channel[i].program = 0;
		Channel[i].volume = 127;
		Channel[i].note = 0;
		Channel[i].detune = 0;
		Channel[i].pan = 64;
	}
	noise_feedback = 0;
	noise_control = 0;
}

void 
DCSGDRV::VgmSet(VGMOUT *p)
{
	pVgmOut = p;
}

void 
DCSGDRV::KeyOn(unsigned int ch)
{
	if (ch >= DCSG_CHANNEL_MAX) return;
	Channel[ch].eg_phase = EG_Attack;
	Channel[ch].key_on = true;
	SetVolume(ch, Channel[ch].volume);
}

void 
DCSGDRV::KeyOff(unsigned int ch)
{
	if (ch >= DCSG_CHANNEL_MAX) return;
	SetReg(0x80 + (volume_slot[ch] << 4) + 0xf);
	Channel[ch].eg_phase = EG_Release;
	Channel[ch].key_on = false;
}

void
DCSGDRV::SetNote(unsigned int ch, unsigned int note)
{
	if (ch >= DCSG_CHANNEL_MAX) return;
	Channel[ch].note = note;

	// noise
	if (ch == 3)
	{
		return;
	}

	int fnum;
	if (note >= (sizeof(DcsgKeyTable) / sizeof(short))) {
		fnum = 0;
	} else {
		int detune = 0;
		if (Channel[ch].detune > 0) {
			detune = Channel[ch].detune * (DcsgKeyTable[note] - DcsgKeyTable[note+12]) / 8191;
		} else {
			detune = Channel[ch].detune * (DcsgKeyTable[note-12] - DcsgKeyTable[note]) / 8191;
		}

		fnum = DcsgKeyTable[note] - detune;
		if (fnum > 0x3ff) fnum = 0x3ff;
		if (fnum < 0) fnum = 0;
	}
	SetReg(0x80 + (freq_slot[ch] << 4) + (fnum & 0xf));
	SetReg(fnum >> 4);
}

void 
DCSGDRV::SetDetune(unsigned int ch, int detune)
{
	if (ch >= DCSG_CHANNEL_MAX) return;
	Channel[ch].detune = detune;
	SetNote(ch, Channel[ch].note);
}

void 
DCSGDRV::SetVolume(unsigned int ch, unsigned int vol)
{
	if (ch >= DCSG_CHANNEL_MAX) return;

	if (vol > 127) vol = 127;
	Channel[ch].volume = vol;

	if (Channel[ch].key_on)
	{
		int v = (0xf - (vol/8));
		SetReg(0x80 + (volume_slot[ch] << 4) + v);
	}
}

void 
DCSGDRV::SetPan(unsigned int ch, unsigned int pan)
{
	return;
}

void 
DCSGDRV::SetProgram(unsigned int ch, unsigned int program)
{
	if (ch == 3) {
		program &= 7;
		SetNoiseFeedback(program >> 2);
		SetNoiseControl(program & 3);
	}
	Channel[ch].program = program;
	return;
}

void 
DCSGDRV::SetNoiseFeedback(unsigned int control)
{
	if (control) 
	{
		noise_feedback = 1;
	}
	else
	{
		noise_feedback = 0;
	}
	SetReg(0x80 + (freq_slot[3] << 4) + (noise_feedback << 2) + noise_control);
	return;
}

void 
DCSGDRV::SetNoiseControl(unsigned int freq)
{
	freq &= 3;
	noise_control = freq;
	SetReg(0x80 + (freq_slot[3] << 4) + (noise_feedback << 2) + noise_control);

	return;
}
