#ifndef DCSGDRV_HPP
#define DCSGDRV_HPP

#include <vector>
#include "vgmout.hpp"

class DCSGDRV {
	static const unsigned int DCSG_CLOCK  = 3579545;
	static const unsigned int DCSG_CHANNEL_MAX = 4;

	enum {
		VOL_MAX = 127,
		PROG_MAX = 127,
	};
	enum {
		EG_Stop = 0,
		EG_Attack,
		EG_Decay,
		EG_Sustain,
		EG_Release,
	};

	struct {
		bool key_on;
		unsigned char program;
		unsigned char pan;
		unsigned char velocity;
		unsigned char volume;
		unsigned char note;
		int detune;
		int eg_phase;
	} Channel[DCSG_CHANNEL_MAX];
	
	int noise_feedback; // 0:mCY, 1:zCgmCY
	int noise_control;  // 00:clk/512
	                    // 01:clk/1024
	                    // 10:clk/2048
	                    // 11:clk (= 3579545)

//	FMInst *pFminst;
	VGMOUT *pVgmOut;
	void SetNoiseFeedback(unsigned int control);
	void SetNoiseControl(unsigned int freq);

public:
	DCSGDRV();
	void VgmSet(VGMOUT *p);
	void KeyOn(unsigned int ch);
	void KeyOff(unsigned int ch);
	void SetNote(unsigned int ch, unsigned int note);
	void SetDetune(unsigned int ch, int detune);
	void SetVolume(unsigned int ch, unsigned int vol);
	void SetPan(unsigned int ch, unsigned int pan);
	void SetProgram(unsigned int ch, unsigned int program);

	void SetReg(unsigned int value);
};

#endif
