#include <boost/throw_exception.hpp>
#include <boost/spirit/core.hpp>
#include <boost/spirit/utility/confix.hpp>
#include <iostream>
#include <stack>
#include "fminst.hpp"
#include "opmreg.hpp"

using namespace std;
using namespace boost::spirit;

//-----------------------------------------------------------------------------
static int CurrentProgNum = 0;
static int errMsk;
static int line;
static FMINST *pFm;
static FMINST *pInst;

//-----------------------------------------------------------------------------
static stack<int> stk;
void push(int n) {
	stk.push(n);
}
int pop() {
	int t = stk.top();
	stk.pop();
	return t;
}
int size() {
	return stk.size();
}
void clear() {
	stk = stack<int>();
}

//-----------------------------------------------------------------------------
void Int(int n)
{
	push(n);
}
void Prog(const char* const, const char* const)
{
	if (size() < 1) {
		fprintf(stderr,".OPM Load Syntax Error : Program\n");
	}
	const int prog = pop();

	CurrentProgNum = prog;
	if (CurrentProgNum < 0)
		CurrentProgNum = 0;
	if (CurrentProgNum > 127)
		CurrentProgNum = 127;
	pInst = &pFm[CurrentProgNum];
//	printf("Prog %d\n", prog);
	clear();
}
void Lfo(const char* const, const char* const)
{
	if (size() != 5) {
		fprintf(stderr,".OPM Load Syntax Error : Lfo\n");
	}
	const int NFRQ = pop();
	const int WF   = pop();
	const int PMD  = pop();
	const int AMD  = pop();
	const int SPD  = pop();
	pInst->NFRQ = NFRQ;
	pInst->WF   = WF  ;
	pInst->PMD  = PMD ;
	pInst->AMD  = AMD ;
	pInst->SPD  = SPD ;
//	printf("LF:%03d %03d %03d %03d %03d\n", SPD,AMD,PMD,WF,NFRQ);
	clear();
}
void Ch(const char* const, const char* const)
{
	if (size() != 7) {
		fprintf(stderr,".OPM Load Syntax Error : Channel\n");
	}
	const int NE   = pop();
	const int SLOT = pop();
	const int PMS  = pop();
	const int AMS  = pop();
	const int CON  = pop();
	const int FB   = pop();
	const int RL   = pop();
	pInst->NE   = NE  ;
	pInst->SLOT = SLOT;
	pInst->PMS  = PMS ;
	pInst->AMS  = AMS ;
	pInst->CON  = CON ;
	pInst->FB   = FB  ;
	pInst->RL   = RL  ;
	pInst->RL >>= OPMREG::RL_SHIFT;
//	printf("// PAN  FL CON AMS PMS SLOT NE\n");
//	printf("CH:%03d %03d %03d %03d %03d %03d %03d \n", RL,FB,CON,AMS,PMS,SLOT,NE);
	clear();
}
void Op(const char* const, const char* const)
{
	if (size() != 12) {
		fprintf(stderr,".OPM Load Syntax Error : Operator\n");
	}
	const int AME = pop();
	const int DT2 = pop();
	const int DT1 = pop();
	const int MUL = pop();
	const int KS  = pop();
	const int TL  = pop();
	const int D1L = pop();
	const int RR  = pop();
	const int D2R = pop();
	const int D1R = pop();
	const int AR  = pop();
	const int op  = pop();
	OP *pOp = &pInst->op[op];
	pOp->AME = AME;
	pOp->AME >>= OPMREG::AME_SHIFT;
	pOp->DT2 = DT2;
	pOp->DT1 = DT1;
	pOp->MUL = MUL;
	pOp->KS  = KS ;
	pOp->TL  = TL ;
	pOp->D1L = D1L;
	pOp->RR  = RR ;
	pOp->D2R = D2R;
	pOp->D1R = D1R;
	pOp->AR  = AR ;
//	printf("Op%d:%03d %03d %03d %03d %03d %03d %03d %03d %03d %03d %03d\n"
//		, op, AR,D1R,D2R,RR,D1L,TL,KS,MUL,DT1,DT2,AME);
	clear();
}
void OpStr(const char* const a, const char* const b)
{
	char Opr[4];
	memcpy(Opr, a, b - a);
	Opr[3] = 0;
	if (strcmp("M1:", Opr) == 0) push(0);
	if (strcmp("C1:", Opr) == 0) push(2);
	if (strcmp("M2:", Opr) == 0) push(1);
	if (strcmp("C2:", Opr) == 0) push(3);
}


//-----------------------------------------------------------------------------
struct comment_parser : public grammar<comment_parser> {
	template<typename S> struct definition {
		rule<S> comment;
		definition(const comment_parser& self) {
			comment =
				comment_p("//")
				| comment_p("/*", "*/")
				;
		}
		const rule<S>& start() const { return comment; }
	};
};

struct opm_reader : public grammar<opm_reader> {
	template<typename S> struct definition {
		rule<S> opm, inst, prog, lfo, ch, op, param;
		definition(const opm_reader& self)
		{
			param = +(int_p[&Int] >> *blank_p);
			prog = str_p("@:")   >> int_p[&Int] >> !(+blank_p >> *(anychar_p - '\n'));
			lfo  = str_p("LFO:") >> *blank_p >> param;
			ch   = str_p("CH:")  >> *blank_p >> param;
			op   = (str_p("M1:") | str_p("C1:") | str_p("M2:") | str_p("C2:"))[&OpStr] >> *blank_p >> param;
			inst = prog[&Prog] 
			       >> *space_p >> lfo[&Lfo] 
			       >> *space_p >> ch[&Ch] 
			       >> *space_p >> +(op[&Op] >> *space_p)
			       >> *space_p;
			opm = *space_p >> *inst >> *space_p;
		}

		const rule<S>& start() const { return opm; }
	};
};


//-----------------------------------------------------------------------------
unsigned int 
FMInst::GetMaxInst(void)
{
	return FM_PROGRAM_MAX;
}

int 
FMInst::Load(char *filename)
{
	FILE *fp;
    string str;
	char strbuf[512];

	pFm = fminst;
	CurrentProgNum = 0;
	errMsk = 0;
	line = 0;

	if ((fp = fopen(filename, "r")) == NULL) {
		return (ERR_NOFILE);
	}
	while(NULL != fgets(strbuf, 512, fp))
	{
		str += strbuf;
	}
	fclose(fp);

	opm_reader opm_read;
	comment_parser comment;
	parse_info<> r = parse(str.c_str(), opm_read, comment);

	if(r.full)
		return (ERR_NOERR);
	else
		return (ERR_SYNTAX);
}
