#include <iostream>
#include <fstream>

#include "midic.hpp"
#include "midip.hpp"
#include "fminst.hpp"
#include "opn2drv.hpp"
#include "dcsgdrv.hpp"
#include "vgmout.hpp"

static const int FRAME_SKIP_MAX = 5;

// static const Uint32 OPM_CLOCK  = 3580000; // 3579545
// FM::OPM opm;

FMInst fminst;

/*
OPN2DRV opn2drv;
MIDI::MIDIP midip;
SDL_Surface *screen;
*/
enum {
	FM_BANK_A = 0,
	FM_BANK_B = 1,
};

const struct {
	char *opm;
	char *mid;
} file[] = {
	{ "./SonicSisters.opm", "smf/cde.mid", },
	{ "./SonicSisters.opm", "smf/SonicSisters_01.mid", },
	{ "./imo_1.opm", "smf/DAWN.MID", },
	{ "./family.opm", "smf/DOORINT.MID", },
	{ "./oke_psd2.opm", "smf/PSGD95.MID", },
	{ "./imo_1.opm", "smf/TECH.MID", },
	{ "./sz2_s1.opm", "smf/sz2_s1.mid", },
//	{ "./1_1459A2.opm", "smf/1_1459A2.MID", },
//	{ "./hanger.opm", "smf/HANGER.MID", },
};

const int file_num = 0;

//==============================================================================

int main(int argc, char **argv)
{
	if (argc < 3) {
		printf("usage: vgmgen.exe hoge.opm hoge.mid\n");
		return 1;
	}
	
//	int errcode = fminst.Load( file[file_num].opm ) ;
	int errcode = fminst.Load( argv[1] ) ;
	if ( errcode ) {
		std::cerr << "OPM file read error (" << errcode << "), abort..." << std::endl;
		return -1;
	}
//	fmdrv.InstSet(&fminst);
//	midip.SetFmdrv(&fmdrv);

	{
		std::ifstream filein;
		filein.open( argv[2] , std::ios::in | std::ios::binary);
		
		if (filein) {
			char *p;

			filein.seekg(0, std::ios::end);
			unsigned int sz = filein.tellg();
			filein.seekg(0, std::ios::beg);

			p = new char[sz];
			filein.read( p, sz );
			if (filein.bad()) {
				std::cerr << "SMF file read error, abort..." << std::endl;
				return -1;
			}
			filein.close();

			MIDI::MIDIC *midic = new MIDI::MIDIC;
			MIDI::DMIDI *dm = midic->convert_smf(p);
			if (!dm) {
				return 1;
			}
			dm = midic->convert_clock_to_frame(dm, 60);

			VGMOUT *vgmout = new VGMOUT;
			OPN2DRV *fmdrv = new OPN2DRV;
			DCSGDRV *dcsgdrv = new DCSGDRV;
			MIDI::MIDIP *midip = new MIDI::MIDIP;

			fmdrv->InstSet(&fminst);
			fmdrv->VgmSet(vgmout);
			dcsgdrv->VgmSet(vgmout);

			midip->SetFmdrv(fmdrv);
			midip->SetDcsgdrv(dcsgdrv);
			midip->SetVgmout(vgmout);
			midip->SetConvertedSmf(dm);

			midip->Render();
			vgmout->Save("test");

			delete midip;
			delete fmdrv;
			delete vgmout;
			delete midic;
		} else {
			std::cerr << "File open error, abort..." << std::endl;
			return -1;
		}
	}

	return 0;
}
