#include "midi.hpp"
#include "midip.hpp"

namespace MIDI {

//--------------------------------------
// 
//--------------------------------------
void 
MIDIP::SetConvertedSmf(DMIDI *dmidi)
{
	this->dmidi = dmidi;
	eventPoint = dmidi->events.begin();
	endPoint = dmidi->events.end();
	loopPoint = dmidi->events.end();
	loopCount = 0;

//	fmdrv->SetTimerA_ByTempo(tempo, dmidi->tpqn);
//	fmdrv->SetReg(REG_IRQ, LOAD_A_MASK | LOAD_B_MASK | IRQEN_A_MASK | IRQEN_B_MASK);
}

void 
MIDIP::SetFmdrv(OPN2DRV* fmdrv)
{
	this->fmdrv = fmdrv;
}
void 
MIDIP::SetDcsgdrv(DCSGDRV* dcsgdrv)
{
	this->dcsgdrv = dcsgdrv;
}

void 
MIDIP::SetVgmout(VGMOUT* vgmout)
{
	this->vgmout = vgmout;
}

void 
MIDIP::AnalyzeControlChange(
	unsigned int ch,
	unsigned int lsb,
	unsigned int msb)
{
	switch (lsb) {
/*
		case BANK_MSB:
		case BANK_LSB:
		case MODULATION:
		case HOLD:
		case PORTAMENT:
		case LEGATO:
		case MONO:
		case POLY:
			break;
*/
		case PORTAMENT_TIME: {
			if (ch < 6)
			{
				fmdrv->SetPortamentTime(ch, msb);
			}
			break;
		}
		case VOLUME: {
			track[ch].exp = msb;
			break;
		}
		case EXPRESSION: {
			track[ch].exp = msb;
			if (ch < 6)
			{
				
			} else
			if (ch < 10)
			{
				dcsgdrv->SetVolume(ch - 6, msb);
			}
			break;
		}
		case PANPOT: {
			track[ch].pan = msb;
			if (ch < 6)
			{
				fmdrv->SetPan(ch, msb);
			}
			break;
		}

		case LOOP_POINT: {
			vgmout->SetLoopHere();
		}

		case DATAENTRY_MSB: {
			if (track[ch].rpn_lsb == 0 && track[ch].rpn_msb == 0) {
				if (msb > 32) msb = 0;
				track[ch].bendrange = msb;
			}
			break;
		}
		case RPN_LSB: {
			track[ch].rpn_lsb = msb;
			break;
		}
		case RPN_MSB: {
			track[ch].rpn_msb = msb;
			break;
		}
	}
}

//--------------------------------------
// ret: none
//--------------------------------------
void 
MIDIP::InterruptTimerA(void)
{
}

//--------------------------------------
// ret: none
//--------------------------------------
void 
MIDIP::InterruptTimerB(void)
{
}

//--------------------------------------
// ret: none
//--------------------------------------
void 
MIDIP::Analyse(std::vector<DMIDI_EVENT>::iterator& e)
{
	// I[}[J[
	if (e->mes == DMIDI_SYS) {
		if (e->lsb == 0xff && e->msb == 0xff) {
			vgmout->WriteEndMarker();
			return;
		}
	}

	// AiCY
	{
		const unsigned int& ch = e->ch;
		const unsigned int& mes = e->mes;
		const unsigned int& lsb = e->lsb;
		const unsigned int& msb = e->msb;

		switch (mes & 0xf0) {
			case DMIDI_SYS: {
				switch (ch | DMIDI_SYS) {
					case DMIDI_SYS_BPM: {
						break;
					}
					case DMIDI_SYS_ENDDATA:
					default: {
						break;
					}
				}
				break;
			}
			
			case PITCHBEND: {
				int pitch = static_cast<short>(lsb | msb << 8);
				if (ch < 6) 
				{
					fmdrv->SetDetune(ch, pitch * track[ch].bendrange / 12 * 0x284 / 8192 );
				} else
				if (ch < 9)
				{
					dcsgdrv->SetDetune(ch - 6, pitch);
				}
				break;
			}
			
			case CC: {
				AnalyzeControlChange(ch, lsb, msb);
				break;
			}

			case PC: {
				if (ch < 6) 
				{
					fmdrv->KeyOff(ch);
					fmdrv->SetProgram(ch, lsb);
					track[ch].key_on = false;
				} else 
				if (ch < 10)
				{
					dcsgdrv->SetProgram(ch - 6, lsb);
				}
				break;
			}

			case NOTE_ON: {
				if (0 == msb) {
					if (track[ch].key_on)
						if (track[ch].note == lsb) {
							track[ch].key_on = false;
							if (ch < 6) 
							{
								fmdrv->KeyOff(ch);
							} else 
							if (ch < 10)
							{
								dcsgdrv->KeyOff(ch - 6);
							}
						}
				} else {
					if (track[ch].key_on) {
						if (ch < 6) 
						{
//							fmdrv->KeyOff(ch);
						} else 
						if (ch < 10)
						{
							dcsgdrv->KeyOff(ch - 6);
						}
					}
					int vel = 97 + (msb * 30) / 127;
					if (ch < 6) 
					{
						fmdrv->SetNote(ch, lsb);
						fmdrv->SetVolume(ch, vel);
						fmdrv->KeyOn(ch);
					} else 
					if (ch < 10)
					{
						dcsgdrv->SetNote(ch - 6, lsb);
						dcsgdrv->KeyOn(ch - 6);
					}
					track[ch].key_on = true;
					track[ch].note = lsb;
					track[ch].velocity = vel;
				}
				break;
			}

			case NOTE_OFF: {
				if (track[ch].key_on) {
					track[ch].key_on = false;
					fmdrv->KeyOff(ch);
				}
				break;
			}
			
			default: {
				break;
			}
		}
	}
}


//--------------------------------------
// ret: none
//--------------------------------------
void 
MIDIP::Render()
{

	std::vector<DMIDI_EVENT>::iterator it = dmidi->events.begin();
	std::vector<DMIDI_EVENT>::iterator end = dmidi->events.end();

	unsigned int last_clock = 0;
	while (it != end)
	{
		while (last_clock != it->clock) 
		{
			vgmout->WriteWait_60();
			fmdrv->RunFrame();
			// dcsgdrv->Envelope();
			last_clock++;
		}
		Analyse(it);
		it++;
	}

}


} // namespace MIDI
