#ifndef MIDIP_HPP
#define MIDIP_HPP

#include <vector>
#include <list>

#include "opn2drv.hpp"
#include "dcsgdrv.hpp"
#include "vgmout.hpp"
#include "midid.hpp"

namespace MIDI
{
static const unsigned int BASE_TEMPO = 120;
static const unsigned int BASE_SAMPLE_RATE = 44100;

class MIDIP
{
private:
	OPN2DRV *fmdrv;
	DCSGDRV *dcsgdrv;
	VGMOUT *vgmout;

	DMIDI *dmidi;
	std::vector<DMIDI_EVENT>::iterator eventPoint;
	std::vector<DMIDI_EVENT>::iterator endPoint;
	std::vector<DMIDI_EVENT>::iterator loopPoint;

	unsigned int clockCount;
	unsigned int loopCount;
	unsigned int sampleRate;
	unsigned int tempo;

	void AnalyzeControlChange(unsigned int ch, unsigned int lsb, unsigned int msb);
	void InterruptTimerA(void);
	void InterruptTimerB(void);
	
	struct TrackStatus {
		unsigned char prog;
		unsigned char vol;
		unsigned char exp;
		unsigned char pan;

		unsigned char bendrange;
		int pitchbend;
		int porta_time;
		int porta_switch;

		int modulation;
		int mod_depth;
		int mod_speed;
		int mod_count;

		unsigned int rpn_msb;
		unsigned int rpn_lsb;

		bool key_on;
		unsigned int note;
		unsigned int velocity;

		struct _note {
			int note;
			int vel;
		};
		std::list<_note> note_;

		TrackStatus() {
			key_on = false;
			note = 0;
			bendrange = 12;
			rpn_msb = 0;
			rpn_lsb = 0;
		}
	} track[32];

public:
	void SetFmdrv(OPN2DRV* fmdrv);
	void SetDcsgdrv(DCSGDRV* dcsgdrv);
	void SetVgmout(VGMOUT* vgmout);

	void SetConvertedSmf(DMIDI *dmidi);
	void Analyse(std::vector<DMIDI_EVENT>::iterator& e);
	void Render();

	MIDIP() {
		sampleRate = BASE_SAMPLE_RATE;
		tempo = BASE_TEMPO;
	}
};

}

#endif
