#ifndef OPMREG_HPP
#define OPMREG_HPP
namespace OPMREG {
	enum {
		BIT0	 = 1,
		BIT1	 = (1<<1),
		BIT2	 = (1<<2),
		BIT3	 = (1<<3),
		BIT4	 = (1<<4),
		BIT5	 = (1<<5),
		BIT6	 = (1<<6),
		BIT7	 = (1<<7),

		REG_KEY		 = 0x08,
		REG_NOISE	 = 0x0F,
		REG_CLOCKA1	 = 0x10,
		REG_CLOCKA2	 = 0x11,
		REG_CLOCKB	 = 0x12,
		REG_IRQ		 = 0x14,
		REG_LFRQ	 = 0x18,
		REG_AMPMD	 = 0x19,
		REG_WFORM	 = 0x1b,

		REG_RL_FB_CON = 0x20,
		REG_KC		 = 0x28,
		REG_KF		 = 0x30,
		REG_PMS_AMS	 = 0x38,
		REG_DT1_MUL	 = 0x40,
		REG_TL		 = 0x60,
		REG_KS_AR	 = 0x80,
		REG_AME_D1R	 = 0xA0,
		REG_DT2_D2R	 = 0xC0,
		REG_D1L_RR	 = 0xE0,

		DETUNE_SHIFT	 = 6,
		DETUNE_MASK		 = 63,

		TIMER_A_FLAG	 = (BIT0),
		TIMER_B_FLAG	 = (BIT1),

		// 08 slot & channel
		SN_MASK			 = (BIT6 | BIT5 | BIT4 | BIT3),
		CHNO_MASK		 = (BIT2 | BIT1 | BIT0),
		// 0F 
		NE_MASK			 = (BIT7),
		NFRQ_MASK		 = (BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
		// 10
		CLKA1_MASK		 = (BIT7 | BIT6 | BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
		// 11
		CLKA2_MASK		 = (BIT1 | BIT0),
		// 12
		CLKB_MASK		 = (BIT7 | BIT6 | BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
		// 14
		CSM_MASK		 = (BIT7),
		CSM_SHIFT		 = 7,
		FRESET_B_MASK	 = (BIT5),
		FRESET_B_SHIFT	 = 5,
		FRESET_A_MASK	 = (BIT4),
		FRESET_A_SHIFT	 = 4,
		IRQEN_B_MASK	 = (BIT3),
		IRQEN_B_SHIFT	 = 3,
		IRQEN_A_MASK	 = (BIT2),
		IRQEN_A_SHIFT	 = 2,
		LOAD_B_MASK		 = (BIT1),
		LOAD_B_SHIFT	 = 1,
		LOAD_A_MASK		 = (BIT0),
		LOAD_A_SHIFT	 = 0,
		// 18
		LFRQ_MASK	 = (BIT7 | BIT6 | BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
		// 19
		PMAMD_MASK	 = (BIT7 | BIT6 | BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
		PM_FLAG		 = BIT7,
		// 1B
		CT_MASK		 = (BIT7 | BIT6),
		CT_SHIFT	 = 6,
		W_MASK		 = (BIT1 | BIT0),
		W_SHIFT		 = 0,
		WAVE_SAW	 = 0,
		WAVE_SQUARE	 = 1,
		WAVE_TRI	 = 2,
		WAVE_NOISE	 = 3,

		// 20-27
		RL_LEFT	 	 = (BIT0),
		RL_RIGHT	 = (BIT1),
		RL_CENTER	 = RL_LEFT | RL_RIGHT,

		RL_MASK		 = (BIT7 | BIT6),
		RL_SHIFT	 = 6,
		FB_MASK		 = (BIT5 | BIT4 | BIT3),
		FB_SHIFT	 = 3,
		CON_MASK	 = (BIT2 | BIT1 | BIT0),
		CON_SHIFT	 = 0,
		// 28-2F
		KC_MASK		 = (BIT6 | BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
		KC_SHIFT	 = 0,
		KEY_MASK	 = (BIT3 | BIT2 | BIT1 | BIT0),
		KEY_SHIFT	 = 0,
		OCT_MASK	 = (BIT6 | BIT5 | BIT4),
		OCT_SHIFT	 = 4,
		// 30-37
		KF_MASK		 = (BIT7 | BIT6 | BIT5 | BIT4 | BIT3 | BIT2),
		KF_SHIFT	 = 2,
		// 38-3F
		PMS_MASK	 = (BIT7 | BIT6 | BIT5),
		PMS_SHIFT	 = 5,
		AMS_MASK	 = (BIT1 | BIT0),
		AMS_SHIFT	 = 0,
		// 40-5F
		DT1_MASK	 = (BIT6 | BIT5 | BIT4),
		DT1_SHIFT	 = 4,
		MUL_MASK	 = (BIT3 | BIT2 | BIT1 | BIT0),
		MUL_SHIFT	 = 0,
		// 60-7F
		TL_MASK		 = (BIT6 | BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
		TL_SHIFT	 = 0,
		// 80-9F
		KS_MASK		 = (BIT7 | BIT6),
		KS_SHIFT	 = 6,
		AR_MASK		 = (BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
		AR_SHIFT	 = 0,
		// A0-BF
		AME_MASK	 = (BIT7),
		AME_SHIFT	 = 7,
		D1R_MASK	 = (BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
		D1R_SHIFT	 = 0,
		// C0-DF
		DT2_MASK	 = (BIT7 | BIT6),
		DT2_SHIFT	 = 6,
		D2R_MASK	 = (BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
		D2R_SHIFT	 = 0,
		// E0-FF
		D1L_MASK	 = (BIT7 | BIT6 | BIT5 | BIT4),
		D1L_SHIFT	 = 4,
		RR_MASK		 = (BIT3 | BIT2 | BIT1 | BIT0),
		RR_SHIFT	 = 0,
	};
}
#endif
