
// ym2612_clock  : 7670453
// opn2 f-number : (144 * freq * 2^20 / clock) / 2^(block-1)

#include "opn2drv.hpp"
#include "opn2reg.hpp"
using namespace OPN2REG;

static const unsigned short Opn2KeyTable[13] = {
	0x0283,0x02aa,0x02d2,0x02fd,0x032b,0x035b,
	0x038e,0x03c4,0x03fd,0x043a,0x047b,0x04bf,
	0x0507,
};

//---------------------------------------------------------------------

void
OPN2DRV::SetReg(unsigned int bank, unsigned int reg, unsigned int value)
{
	if (pVgmOut) {
		pVgmOut->WriteRegOpn2(bank, reg, value);
	}
}

unsigned int 
get_fm_channel(unsigned int ch)
{
	return (ch < 3) ? (ch & 3) : (ch + 1);
}

unsigned int 
get_fm_bank(unsigned int ch)
{
	return (ch < 3) ? FM_BANK_A : FM_BANK_B;
}

unsigned int 
get_fm_slot(unsigned int slot)
{
	const unsigned int slot_table[] = { 0, 0x8, 0x4, 0xC, };
	return (slot < 4) ? slot_table[slot] : 0;
}

//---------------------------------------------------------------------

OPN2DRV::OPN2DRV()
{
	pVgmOut = NULL;
	for (unsigned int i = 0; i < FM_CHANNEL_MAX; i++) {
		Channel[i].key_on = false;
		Channel[i].program = 0;
		Channel[i].volume = 127;
		Channel[i].note = 0;
		Channel[i].note_prev = 0;
		Channel[i].detune = 0;
		Channel[i].pan = 64;
		Channel[i].porta_time = 0;
		Channel[i].porta_detune = 0;
		Channel[i].porta_frame = 0;
		Channel[i].porta_delta = 0;
	}
}

void 
OPN2DRV::VgmSet(VGMOUT *p)
{
	pVgmOut = p;
	SetReg(FM_BANK_A, REG_TIMERA_H, 0x00);
	SetReg(FM_BANK_A, REG_TIMERA_L, 0x00);
	SetReg(FM_BANK_A, REG_TIMERB, 0x00);
	SetReg(FM_BANK_A, REG_TIMER_CNT, 0x3f);

	/* LFOݒ  LFO=ON, Freq=6.02Hz */
	//0  	1 	 	2 	 	3 	 	4	  	5	  	6 	 	7		8
	//3.98 	5.56 	6.02 	6.37 	6.88 	9.63 	48.1 	72.2	Enabled
	SetReg(FM_BANK_A, REG_LFO, 0x0a);

}

void 
OPN2DRV::InstSet(FMInst* fminst)
{
	pFminst = fminst;
}

void 
OPN2DRV::KeyOn(unsigned int ch)
{
	if (ch >= FM_CHANNEL_MAX) return;
	SetReg(FM_BANK_A, REG_KEY, SLOT_MASK + get_fm_channel(ch));
	Channel[ch].key_on = true;

	if (Channel[ch].porta_time)
	{
		if (Channel[ch].note_prev != Channel[ch].note)
		{
			int prev = Opn2KeyTable[(Channel[ch].note_prev % 12)] 
				+ (0x0283 * (Channel[ch].note_prev / 12));
			int now = Opn2KeyTable[(Channel[ch].note % 12)] 
				+ (0x0283 * (Channel[ch].note / 12));
			Channel[ch].porta_detune = prev - now;
			Channel[ch].porta_delta = Channel[ch].porta_detune / Channel[ch].porta_time;
			Channel[ch].porta_frame = Channel[ch].porta_time;
			SetNote(ch, Channel[ch].note);
		} else {
			Channel[ch].porta_detune = 0;
			Channel[ch].porta_delta = 0;
			Channel[ch].porta_frame = 0;
		}
	}
	Channel[ch].note_prev = Channel[ch].note;

//	if (pVgmOut) {
//		pVgmOut->PtrintAllRegister();
//	}

}

void 
OPN2DRV::KeyOff(unsigned int ch)
{
	if (ch >= FM_CHANNEL_MAX) return;
	SetReg(FM_BANK_A, REG_KEY, get_fm_channel(ch));
	Channel[ch].key_on = false;
}

void 
OPN2DRV::SetNote(unsigned int ch, unsigned int note)
{
	if (ch >= FM_CHANNEL_MAX) return;
	Channel[ch].note = note;

	int block = (note / 12);
	int fnumber = Opn2KeyTable[(note % 12)] + Channel[ch].detune + Channel[ch].porta_detune;
	while (fnumber < Opn2KeyTable[0])
	{
		block--;
		fnumber += Opn2KeyTable[0];
	}
	while (fnumber > Opn2KeyTable[12])
	{
		block++;
		fnumber -= Opn2KeyTable[0];
	}
	if (block < 0) block = 0;
	if (block > 7) block = 7;

	int fn1 = fnumber & 0xff;
	int fn2 = (fnumber >> 8) | (block << 3);
	int bank = get_fm_bank(ch);
	ch = get_fm_channel(ch) & 3;

	SetReg(bank, REG_FNUMBER_H + (ch & 3), fn2); // ubNɐݒ肷
	SetReg(bank, REG_FNUMBER_L + (ch & 3), fn1);
}

void 
OPN2DRV::SetDetune(unsigned int ch, int detune)
{
	if (ch >= FM_CHANNEL_MAX) return;
	Channel[ch].detune = detune;
	SetNote(ch, Channel[ch].note);
}

void 
OPN2DRV::SetTL(unsigned int ch, unsigned int op, unsigned int vol)
{
	const FMINST& inst = pFminst->fminst[Channel[ch].program];
	const OP& inst_op = inst.op[op];
	int bank = get_fm_bank(ch);
	ch = get_fm_channel(ch) & 3;
	int addr = ch + op * 4;
	int tl = 127 - ((127 - inst_op.TL) * vol / 127);
	SetReg(bank, REG_TL + addr, tl);
}

void 
OPN2DRV::SetVolume(unsigned int ch, unsigned int vol)
{
	if (ch >= FM_CHANNEL_MAX) return;
	if (vol > 127) vol = 127;
	
	const FMINST& inst = pFminst->fminst[Channel[ch].program];

	switch (inst.CON) {
		case 7: 
			SetTL(ch, 0, vol); // op1
		case 6:
		case 5: 
			SetTL(ch, 1, vol); // op2
		case 4: 
			SetTL(ch, 2, vol); // op3
		default:
			SetTL(ch, 3, vol); // op4
			break;
	}
}

void 
OPN2DRV::SetPan(unsigned int ch, unsigned int pan)
{
	if (ch >= FM_CHANNEL_MAX) return;

	Channel[ch].pan = pan;

	int rl;
	if (pan < 64) {
		rl = L_MASK;
	} else
	if (pan == 64) {
		rl = LR_MASK;
	} else {
		rl = R_MASK;
	}
	unsigned int bank = get_fm_bank(ch);
	unsigned int data;
	
	const FMINST& inst = pFminst->fminst[Channel[ch].program];
	ch = get_fm_channel(ch) & 3;
	data = rl
		 | ((inst.AMS << AMS_SHIFT) & AMS_MASK)
		 | ((inst.PMS << PMS_SHIFT) & PMS_MASK);
	SetReg(bank, REG_LR_AMS_PMS + ch, data);
}

void 
OPN2DRV::SetPortamentTime(unsigned int ch, unsigned int porta_time)
{
	if (ch >= FM_CHANNEL_MAX) return;
	Channel[ch].porta_time = porta_time;
}


void 
OPN2DRV::SetTimerA(unsigned short na)
{
	na &= 0x3ff;
	unsigned int msb = (na >> 2) & 0xff;
	unsigned int lsb = na & 0x03;
	SetReg(FM_BANK_A, REG_TIMERA_H, msb);
	SetReg(FM_BANK_A, REG_TIMERA_L, lsb);
}

void 
OPN2DRV::SetTimerB(unsigned char clkb)
{
	SetReg(FM_BANK_A, REG_TIMERB, clkb);
}

void 
OPN2DRV::SetTimerA_ByTempo(unsigned short tempo, unsigned short tpqn)
{
	double timing = 60000.0 / (static_cast<double>(tempo * tpqn) );
	unsigned int na = static_cast<unsigned int>(1024.0 - timing * (static_cast<double>(OPN2_CLOCK) / 1000.0) / 64.0);
	SetTimerA(na);
}

void 
OPN2DRV::SetTimerB_ByTempo(unsigned short tempo, unsigned short tpqn)
{
	double timing = 60 * 1000 / (tempo * tpqn);
	unsigned int nb = static_cast<unsigned int>(256L - timing * (static_cast<double>(OPN2_CLOCK) / 1000L) / 1024L);
	SetTimerB(nb);
}

void 
OPN2DRV::SetTimerB_ByMilliSeconds(unsigned short ms)
{
	unsigned int nb = 256 - (ms * OPN2_CLOCK / 1024);
	SetTimerB(nb);
}

void 
OPN2DRV::RunFrame(void)
{
	for (unsigned int ch = 0; ch < FM_CHANNEL_MAX; ++ch)
	{
		if (Channel[ch].porta_frame)
		{
			Channel[ch].porta_detune -= Channel[ch].porta_delta;
			Channel[ch].porta_frame--;
			if (0 == Channel[ch].porta_frame)
			{
				Channel[ch].porta_detune = 0;
			}
			SetNote(ch, Channel[ch].note);
		}
	}
}

void 
OPN2DRV::SetProgram(unsigned int ch, unsigned int program)
{
	if (ch >= FM_CHANNEL_MAX) return;
	if (program >= pFminst->GetMaxInst()) return;
	
	Channel[ch].program = program;
	const FMINST& inst = pFminst->fminst[program];

	int rl;
	if (Channel[ch].pan < 64) {
		rl = L_MASK;
	} else
	if (Channel[ch].pan == 64) {
		rl = LR_MASK;
	} else {
		rl = R_MASK;
	}

	unsigned int data;
	unsigned int bank = get_fm_bank(ch);
	ch = get_fm_channel(ch) & 3;

	data = ((inst.FB << FB_SHIFT) & FB_MASK)
		 | ((inst.CON << CON_SHIFT) & CON_MASK);
	SetReg(bank, REG_FB_CON + ch, data);

	data = rl
		 | ((inst.AMS << AMS_SHIFT) & AMS_MASK)
		 | ((inst.PMS << PMS_SHIFT) & PMS_MASK);
	SetReg(bank, REG_LR_AMS_PMS + ch, data);
/*	printf("CON  FB AMS PMS LR\n");
	printf("%03d %03d %03d %03d %03d\n",
		inst.CON,
		inst.FB,
		inst.AMS,
		inst.PMS,
		inst.RL
	);
*/
//	printf(" AR  DR  SR  RR  SL  TL  KS  ML  DT AMS\n");

	for (int op = 0; op < 4; op++) {
		const OP& inst_op = inst.op[op];
		int addr = ch + (op << 2);
/*
		printf("%03d %03d %03d %03d %03d %03d %03d %03d %03d %03d\n",
			inst_op.AR ,
			inst_op.D1R ,
			inst_op. D2R ,
			inst_op.RR ,
			inst_op.D1L ,
			inst_op.TL ,
			inst_op.KS ,
			inst_op.MUL ,
			inst_op.DT1 ,
			inst_op.AME 
		);
*/
		data = ((inst_op.DT1 << DT_SHIFT) & DT_MASK)
			 | ((inst_op.MUL << MUL_SHIFT) & MUL_MASK);
		SetReg(bank, REG_DT_MUL + addr, data);

		data = ((inst_op.TL << TL_SHIFT) & TL_MASK);
		SetReg(bank, REG_TL + addr, data);

		data = ((inst_op.KS << KS_SHIFT) & KS_MASK)
			 | ((inst_op.AR << AR_SHIFT) & AR_MASK);
		SetReg(bank, REG_KS_AR + addr, data);

		data = ((inst_op.AME << AMON_SHIFT) & AMON_MASK)
			 | ((inst_op.D1R << DR_SHIFT) & DR_MASK);
		SetReg(bank, REG_AMON_DR + addr, data);

		data = ((inst_op.D2R << SR_SHIFT) & SR_MASK);
		SetReg(bank, REG_SR + addr, data);

		data = ((inst_op.D1L << SL_SHIFT) & SL_MASK)
			 | ((inst_op.RR  << RR_SHIFT) & RR_MASK);
		SetReg(bank, REG_SL_RR + addr, data);
	}

	return;
}
