#ifndef OPN2REG_HPP
#define OPN2REG_HPP

namespace OPN2REG {
	enum {
		BIT0	 = 1,
		BIT1	 = (1<<1),
		BIT2	 = (1<<2),
		BIT3	 = (1<<3),
		BIT4	 = (1<<4),
		BIT5	 = (1<<5),
		BIT6	 = (1<<6),
		BIT7	 = (1<<7),

		REG_LFO			 = 0x22,
		REG_TIMERA		 = 0x24, // tA(usec) = 144 * (1024-NA)/clock
		REG_TIMERA_H	 = 0x24, // higher 8bit
		REG_TIMERA_L	 = 0x25, // lower 2bit
		REG_TIMERB		 = 0x26, // tB(usec) = 2304 * (256-NB)/clock
		REG_TIMER_CNT	 = 0x27, // Timer Control
								 // 	D0: Timer A Load (1 = Start)
								 // 	D1: Timer B Load 
								 // 	D2: Status A Enable (1 = Flag on when flow)
								 // 	D3: Status B Enable
								 // 	D4: Timer A Reset (1 = Reset timer & flag)
								 // 	D5: Timer B Reset
								 // 	D6-7: FM Mode
								 // 		0 = Normal
								 // 		1 = FM3Extend
								 // 		2 = CSM
		REG_KEY			 = 0x28, // Key on/off
								 // 	D0-2: Channel Select
								 // 		0 = ch.1
								 // 		1 = ch.2
								 // 		2 = ch.3
								 // 		4 = ch.4
								 // 		5 = ch.5
								 // 		6 = ch.6
								 // 	D4-6: Slot Select
								 // 		D4 = Slot1
								 // 		D5 = Slot2
								 // 		D6 = Slot3
								 // 		D7 = Slot4
		REG_DAC_DATA	 = 0x2A, // bit0 LSB - bit7 MSB
		REG_DAC_SELECT	 = 0x2B, // D7: DAC Enable
		REG_TEST		 = 0x2C,

		// Bank Select
		FM_BANK_A = 0,
		FM_BANK_B = 1,

		// n3,n7,nB,nF͖
		REG_DT_MUL		 = 0x30, // D0-4:MUL  D5-7:DT
		REG_TL			 = 0x40, // D0-7
		REG_KS_AR		 = 0x50, // D0-5:AR   D6-7:KS
		REG_AMON_DR		 = 0x60, // D0-5:DR   D7:AMON
		REG_SR			 = 0x70, // D0-5:SR
		REG_SL_RR		 = 0x80, // D0-4:RR   D5-7:SL
		REG_SSGEG		 = 0x90, // D0-4:SSG-EG

		REG_FNUMBER		 = 0xA0, // 
		REG_FNUMBER_L	 = 0xA0, // D0-7:FnumLSB
		REG_FNUMBER_H	 = 0xA4, // D0-2:FnumMSB D3-5:Block
		REG_FM3_FNUM	 = 0xA8, // 
		REG_FM3_FNUM_L	 = 0xA8, // 
		REG_FM3_FNUM_H	 = 0xAC, // 
		REG_FB_CON		 = 0xB0, // D0-3:CON  D4-6:FB
		REG_LR_AMS_PMS	 = 0xB4, // D0-2:PMS  D4-5:AMS  D6:R  D7:L
		                         // PMS 0: 0
		                         //     1: +-3.4cent
		                         //     2: +-6.7cent
		                         //     3: +-10cent
		                         //     4: +-14cent
		                         //     5: +-20cent
		                         //     6: +-40cent
		                         //     7: +-80cent
		                         // AMS 0: 0
		                         //     1: 1.4dB
		                         //     2: 5.9dB
		                         //     3: 11.8dB

		SLOT_SHIFT		 = 4,
		SLOT_MASK		 = (BIT7 | BIT6 | BIT5 | BIT4),
		CHANNEL_SHIFT	 = 0,
		CHANNEL_MASK	 = (BIT2 | BIT1 | BIT0),

//--------------------------------------------------------

	//	(BIT7 | BIT6 | BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
	// $30
			DT_SHIFT	 = 5,
			DT_MASK		 = (BIT7 | BIT6 | BIT5),
			MUL_SHIFT	 = 0,
			MUL_MASK	 = (BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
	// $40
			TL_SHIFT	 = 0,
			TL_MASK		 = (BIT7 | BIT6 | BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
	// $50
			KS_SHIFT	 = 6,
			KS_MASK		 = (BIT7 | BIT6),
			AR_SHIFT	 = 0,
			AR_MASK		 = (BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
	// $60
			AMON_SHIFT	 = 7,
			AMON_MASK	 = (BIT7),
			DR_SHIFT	 = 0,
			DR_MASK		 = (BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
	// $70
			SR_SHIFT	 = 0,
			SR_MASK		 = (BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
	// $80
			SL_SHIFT	 = 5,
			SL_MASK		 = (BIT7 | BIT6 | BIT5),
			RR_SHIFT	 = 0,
			RR_MASK		 = (BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
	// $90
			SSGEG_SHIFT	 = 0,
			SSGEG_MASK	 = (BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
	// $a0, $a8(FM3)
			FNUM_L_SHIFT = 0,
			FNUM_L_MASK	 = (BIT7 | BIT6 | BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0),
	// $a4, $ac(FM3)
			FNUM_H_SHIFT = 0,
			FNUM_H_MASK	 = (BIT2 | BIT1 | BIT0),
			BLOCK_SHIFT	 = 3,
			BLOCK_MASK	 = (BIT5 | BIT4 | BIT3),
	// $b0
			FB_SHIFT	 = 3,
			FB_MASK		 = (BIT5 | BIT4 | BIT3),
			CON_SHIFT	 = 0,
			CON_MASK	 = (BIT2 | BIT1 | BIT0),
	// $b4
			L_SHIFT		 = 7,
			L_MASK		 = (BIT7),
			R_SHIFT		 = 6,
			R_MASK		 = (BIT6),
			LR_SHIFT	 = 6,
			LR_MASK		 = (BIT7 | BIT6),
			AMS_SHIFT	 = 4,
			AMS_MASK	 = (BIT5 | BIT4),
			PMS_SHIFT	 = 0,
			PMS_MASK	 = (BIT2 | BIT1 | BIT0),
	};
}
#endif
