#include <iostream>
#include <zlib/zlib.h>
#include "vgmout.hpp"

static unsigned char fm_block[2][256] = { {0},{0},};

void
VGMOUT::WriteRegOpn2(unsigned int bank, unsigned int reg, unsigned int value)
{
	bank &= 1;
	reg &= 0xff;
	value &= 0xff;
	unsigned int mes = bank ? 0x53 : 0x52;
	vgm_data.push_back(mes);
	vgm_data.push_back(static_cast<unsigned char>(reg));
	vgm_data.push_back(static_cast<unsigned char>(value));
	::fm_block[bank][reg] = value;
//	printf("%8d: %02X  %02X  %02X\n", total_samples, mes, reg, value);
}

void 
VGMOUT::WriteRegDcsg(unsigned int value)
{
	vgm_data.push_back(0x50);
	vgm_data.push_back(static_cast<unsigned char>(value));
//	printf("%8d: %02X\n", total_samples, value);
}

void 
VGMOUT::WriteWait(unsigned int samples)
{
	while (samples > 65535) {
		vgm_data.push_back(0x61);
		vgm_data.push_back(0xff);
		vgm_data.push_back(0xff);
		samples -= 65535;
		total_samples += 65535;
	}
	vgm_data.push_back(0x61);
	vgm_data.push_back(samples & 0xff);
	vgm_data.push_back((samples >> 8) & 0xff);
	total_samples += samples;
}
void 
VGMOUT::WriteWait_60(void)
{
	vgm_data.push_back(0x62);
	total_samples += 735;
}
void 
VGMOUT::WriteWait_50(void)
{
	vgm_data.push_back(0x63);
	total_samples += 882;
}

void 
VGMOUT::SetLoopHere(void)
{
	loop_point = vgm_data.size();
	loop_point_samples = total_samples;
}

void 
VGMOUT::WriteEndMarker(void)
{
	vgm_data.push_back(0x66);
}

void 
VGMOUT::Save(const char *filename)
{
	size_t sz = sizeof(VGM_HEADER) + vgm_data.size();
	char* data = new char[ sz ];

	vgm_header->eof_offset = 0x40 - 4 + vgm_data.size();
	vgm_header->total_nsamples = total_samples;
	if (loop_point)
	{
		vgm_header->loop_offset = loop_point + 0x40 - 0x1c;
		vgm_header->loop_nsamples = vgm_header->total_nsamples - loop_point_samples;
	}
	vgm_header->vgm_data_offset = 0;

	memcpy( data, vgm_header, 0x40 );
	memcpy( data+0x40, &vgm_data[0], vgm_data.size() );

	FILE* fp = std::fopen("test.vgm", "wb");
	fwrite(data, sizeof(char), sz, fp);
	std::fclose( fp );

	gzFile gzfp = gzopen("test.vgz", "wb");
	gzwrite(gzfp, data, sz);
	gzclose( gzfp );

	delete data;
}

void 
VGMOUT::PtrintAllRegister(void)
{
	for (int bank = 0; bank < 2; ++bank)
	{
		printf("== BANK %d ==\n", bank);
		printf("       0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F\n");
		for (int addr = 0; addr < 256; ++addr)
		{
			if ((addr & 0xf) == 0)
				printf("%02X : ", addr);
			printf("%03d ", ::fm_block[bank][addr] );
			if ((addr & 0xf) == 0xf)
				printf("\n");
		}
	}
	printf("\n");
}
