#ifndef VGMOUT_HPP
#define VGMOUT_HPP

#include <vector>

struct VGM_HEADER {
	char ident[4];
	unsigned int eof_offset;
	unsigned int version;
	unsigned int sn76489_clock;
	unsigned int ym2413_clock;
	unsigned int gd3_offset;
	unsigned int total_nsamples;
	unsigned int loop_offset;
	unsigned int loop_nsamples;
	unsigned int rate;
	unsigned short sn_feedback;
	unsigned char sn_width;
	unsigned char reserved1;
	unsigned int ym2612_clock;
	unsigned int ym2151_clock;
	unsigned int vgm_data_offset;
	unsigned int reserved2;
	unsigned int reserved3;
};

class VGMOUT
{
private:
	unsigned int total_samples;
	unsigned int total_frames;
	unsigned int loop_point;
	unsigned int loop_point_samples;
	VGM_HEADER* vgm_header __attribute__ ((packed));
	std::vector<unsigned char> vgm_data;
public:
	VGMOUT() { 
		total_samples = 0;
		total_frames = 0;
		loop_point = 0;
		loop_point_samples = 0;

		vgm_header = new VGM_HEADER;

		vgm_header->ident[0] = 'V';
		vgm_header->ident[1] = 'g';
		vgm_header->ident[2] = 'm';
		vgm_header->ident[3] = ' ';
		vgm_header->eof_offset = 0;			// must be fixed later
		vgm_header->version = 0x00000150;
		vgm_header->sn76489_clock = 3579545;
		vgm_header->ym2413_clock = 0;
		vgm_header->gd3_offset = 0;
		vgm_header->total_nsamples = 0;		// must be fixed
		vgm_header->loop_offset = 0;		// must be fixed
		vgm_header->loop_nsamples = 0;		// must be fixed
		vgm_header->rate = 60;
		vgm_header->sn_feedback = 0x0009;
		vgm_header->sn_width = 16;
		vgm_header->reserved1 = 0;
		vgm_header->ym2612_clock = 7670453;
		vgm_header->ym2151_clock = 0;
		vgm_header->vgm_data_offset = 0;	// must be fixed (for pcm)
		vgm_header->reserved2 = 0;
		vgm_header->reserved3 = 0;
		vgm_data.clear();
	}
	~VGMOUT() { 
		delete vgm_header;
	}

	void WriteRegOpn2(unsigned int bank, unsigned int reg, unsigned int value);
	void WriteRegDcsg(unsigned int value);
	void WriteWait(unsigned int samples);
	void WriteWait_60(void);
	void WriteWait_50(void);
	void SetLoopHere(void);
	void WriteEndMarker(void);
	void Save(const char *);
	void PtrintAllRegister(void);
};

#endif
